/* -*-mode:C;c-basic-offset:8-*-
 *      Copyright (C) 1994, 1995  Claus-Justus Heine.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 $Source: /homes/cvs/zftape/zftape-buffers.c,v $
 $Author: claus $
 *
 $Revision: 1.20 $
 $Date: 1996/08/31 00:48:53 $
 $State: Exp $
 *
 */

static char RCSid[] =
"$Id: zftape-buffers.c,v 1.20 1996/08/31 00:48:53 claus Exp $";

#include "ftape.h"
#include <linux/errno.h>   /* so what ?    */
#include <linux/string.h>  /* memcpy etc. */
#include <asm/dma.h>       /* MAX_DMA_ADDRESS for __get_dma_pages() macro */
#include <linux/malloc.h>  /* __get_dma_pages() and kmalloc() */
#include <linux/fs.h>      /* for kdev_t */

#include "kernel-interface.h" /* for bits of minor dev. no.      */
#include "ftape-rw.h"         /* for the dma-buffer super-struct */
#include "qic80-compress.h"   /* for the compression buffers     */
#include "qic80-vtbl.h"       /* for the internal vtbl table     */
#include "zftape-buffers.h"   /* for SIZEOF macro                */
#include "ftape-bsm.h"        /* for bad sector map              */
#include "ftape-io.h"         /* for ftape_sleep()               */



/* the dma-buffer is allocated when the tape devices are opened and 
 * released when they are closed. Also, __get_dma_pages() is used 
 * which produces already aligned buffers. In addition, a couple of
 * other buffers is allocated when ftape_open() is called and released
 * on close.
 * Idea for the "allocate on open() and release on close()" scheme was 
 * by Jon Tombs (jon@gte.esi.us.es)
 *
 *
 * Thanks to Jon Tombs (jon@gte.esi.us.es)
 * We use get_dma_pages() instead of kmalloc which gives us the entire memory we
 * need and does not consume 24 bytes for bookkeeping purposes (kmalloc did)
 * The dynamic memory allocation is a lot easier now.
 */

/* 
 * global variables
 */

#ifdef ZFTAPE_DMA_MEM_REDUCE_HACK
/* pointer to the one and only dma buffer. F**in' dma memory allocation */
byte *dma_buffer = NULL;
byte *dma_ptr = NULL;
int dma_segment = -1;
#endif

/* the number of 32kb dma-buffers currently allocated 
 */
int ftape_num_buffers = 0;

/* array of pointers to the dma-buffers. MAX_DMA_BUFFERS can be set in the 
 * Makefile 
 */
static byte *tape_buffers[ MAX_DMA_BUFFERS ] = { NULL, }; 

/* the user may change this variable with  ftape_change_dma_size()
 */
static int desired_num_buffers = NR_BUFFERS;

/*
 *    use this structure to allocate some buffers that are only needed during
 *    tape operation. This spares about 100k of RAM that would be unneccessary
 *    allocated
 *
 *    array of pointers to the pointers to the arrays
 */
#ifndef NO_TRACE_AT_ALL
#define DYN_MEM_ENTRY(a,b,c,d) a, b, c, d
#else
#define DYN_MEM_ENTRY(a,b,c,d) a, b, c
#endif
static struct
{ 
	unsigned long size;   /* size of the buffer to allocate */
	void **location;      /* pointer to the pointer to the buffer */
	kdev_t alloc_mask;    /* allocate buffer only if minor device number 
			       * contains a bit set here 
			       */
#ifndef NO_TRACE_AT_ALL
	char *name;
#endif
} dyn_mem_info[] = {
	{ DYN_MEM_ENTRY(SIZEOF(scratch_buffer), 
			(void **)&scratch_buffer, 
			ZFTAPE_ZIP_MODE, 
			"scratch_buffer") },
	{ DYN_MEM_ENTRY(SIZEOF(deblock_buffer), 
			(void **)&deblock_buffer, 
			0, 
			"deblock_buffer") },
	{ DYN_MEM_ENTRY(SIZEOF(zftape_compression_buffer), 
			(void **)&zftape_compression_buffer, 
			ZFTAPE_ZIP_MODE, 
			"ftape_compression_buffer") },
	{ DYN_MEM_ENTRY(SIZEOF(zftape_compression_wrk_mem), 
			(void **)&zftape_compression_wrk_mem, 
			ZFTAPE_ZIP_MODE, 
			"ftape_compression_wrk_mem") },
	{ DYN_MEM_ENTRY(SIZEOF(buffer), 
			(void **)&buffer, 
			0, 
			"buffer") }
};
#define NR_DYN_MEM NR_ITEMS(dyn_mem_info)

/* Pure 2^n version of get_order */
static inline int __get_order(unsigned long size)
{
	int order;

	size = (size + PAGE_SIZE -1) >> PAGE_SHIFT;
	order = -1;
	do {
		size >>= 1;
		order++;
	} while (size);

	return order;
}

static byte *dma_mem_alloc(int size)
{
	return (byte *)__get_dma_pages(GFP_KERNEL, __get_order( size ) );
}

static int grow_number_of_buffers(int num_buffers);
static int shrink_number_of_buffers(int num_buffers);

/*   called by  init_module(). It computes the order parameter for 
 *   __get_dma_pages()
 */
int
zftape_init_mem( void )
{
	TRACE_FUN( 5, "ftape_init_mem");
	int i;
	
	for ( i = 0 ; i < NR_DYN_MEM ; i++ ) {
		*dyn_mem_info[i].location = NULL;
	}
	/*
	 * Back to reality. It is REALLY unsafe and discouraged to allocate
	 * the dma buffers at open time. All make-linux-swap-hacks I tried
	 * either didn't work or kill processes under certain circumstances.
	 *
	 * Therefore: 
	 */
#ifndef  ALLOC_DMA_BUFFER_WHEN_OPENING_DEVICE
	/* we allocate the minimum amount of dma memory that is needed 
	 * for proper operation. The drivers no longer releases this 
	 * memory when the vfs calles the close routine. 
	 *
	 * But: if the dma memory was increased using MTSETDRVBUFFER,
	 * the release routine decreases the memory down to
	 * MIN_DMA_BUFFERS. Well. Here we are:
	 *
	 */
	if (grow_number_of_buffers(MIN_DMA_BUFFERS) < MIN_DMA_BUFFERS) {
		/* we loose. no further comment.
		 */
		(void)shrink_number_of_buffers(0);
		TRACE_EXIT;
		return -ENOMEM;
	}
#endif
	TRACE_EXIT;
	return 0;
}

/*   there are some more buffers that are allocated on demand. 
 *   cleanup_module() calles this function to be sure to have
 *   released them
 */
void
zftape_uninit_mem( void )
{
	TRACE_FUN( 5, "ftape_uninit_mem");
	zftape_release_compression_map();
	zftape_release_bad_sector_map();
	zftape_release_qic80_vtbl();
#ifndef  ALLOC_DMA_BUFFER_WHEN_OPENING_DEVICE
	(void)shrink_number_of_buffers(0);
#endif
	TRACE_EXIT;
	return;
}

/*
 * this function is called by ftape_zap_read_buffers() to initialize
 * the global variable buffer, the dma buffer supervisor structure
 * This must be called at least once every time the device is opened
 * because the dma buffer is released on close (for kernels > 1.3.15)
 */
void
zftape_init_tape_buffer ( void )
{
	TRACE_FUN( 5, "zftape_init_tape_buffer");
	int i;
	
	for ( i = 0; i < ftape_num_buffers; i++ ) {
		buffer[ i].address = tape_buffers[i];
	}
	TRACE_EXIT;
}

/*
 *  called by read_header_segments() after we know the size of the tape
 *  currently inserted. The compression map would occupy up to about 60kb
 *  for wide "qic" tapes
 */
int zftape_allocate_compression_map( void )
{
	TRACE_FUN( 5, "zftape_allocate_compression_map");
	int alloc_size =   zftape_compression_map_length
		         * sizeof(CompressionMapEntry);

	zftape_release_compression_map();
	if ( zftape_compression_map_length != 0 ) {
		zftape_compression_map = vmalloc(  alloc_size );
		if ( zftape_compression_map != NULL ) {
			TRACEx1( 4, "allocated compression-map, %d bytes", 
				 alloc_size );
			memset( zftape_compression_map, '\0', alloc_size );
			TRACE_EXIT;
			return 0;
		} else {
			TRACE_EXIT;
			return -ENOMEM;
		}
	}
	TRACE_EXIT;
	return 0;  
}
void zftape_release_compression_map( void )
{
	TRACE_FUN( 5, "zftape_release_compression_map");

	if ( zftape_compression_map != NULL ) {
		TRACEx1( 4, "freed compression-map, %d bytes", 
			   zftape_compression_map_length
			 * sizeof(CompressionMapEntry));
		vfree( zftape_compression_map );
		zftape_compression_map = NULL;
	}
	TRACE_EXIT;
}

/*
 *   allocate the bad_sector_map
 */
int zftape_allocate_bad_sector_map( void )
{
	TRACE_FUN( 5, "zftape_allocate_bad_sector_map");
	if ( bad_sector_map == NULL ) {
		bad_sector_map = vmalloc( SIZEOF(bad_sector_map) ); 
		if ( bad_sector_map != NULL ) {
			TRACEx1( 4, "allocated bad_sector_map, %d bytes", 
				 SIZEOF(bad_sector_map) );
			memset( bad_sector_map, '\0', SIZEOF(bad_sector_map) );
			TRACE_EXIT;
			return 0;
		} else {
			TRACE_EXIT;
			return -ENOMEM;
		} 
	} else {
		memset( bad_sector_map, '\0', SIZEOF(bad_sector_map) );
		TRACE_EXIT;
		return 0;
	}
}
void zftape_release_bad_sector_map( void )
{
	TRACE_FUN( 5, "zftape_bad_sector_map");

	if ( bad_sector_map != NULL ) {
		vfree( bad_sector_map );
		TRACEx1( 4, "freed bad_sector_map, %d bytes", 
			 SIZEOF(bad_sector_map) );
		bad_sector_map = NULL;
	}
	TRACE_EXIT;
}

/*
 *   allocate the qic80_internal_vtbl
 */
int zftape_allocate_qic80_vtbl( void )
{
	TRACE_FUN( 5, "zftape_allocate_qic80_vtbl");
	if ( __qic80_vtbl == NULL ) {
		__qic80_vtbl = vmalloc( SIZEOF(__qic80_vtbl) ); 
		if ( __qic80_vtbl != NULL ) {
			TRACEx1( 4, "allocated __qic80_vtbl, %d bytes", 
				 SIZEOF(__qic80_vtbl) );
			memset( __qic80_vtbl, '\0', SIZEOF(__qic80_vtbl) );
			TRACE_EXIT;
			return 0;
		} else {
			TRACE_EXIT;
			return -ENOMEM;
		} 
	} else {
		memset( __qic80_vtbl, '\0', SIZEOF(__qic80_vtbl) );
		TRACE_EXIT;
		return 0;
	}
}
void zftape_release_qic80_vtbl( void )
{
	TRACE_FUN( 5, "zftape_release_qic80_vtbl");
	if ( __qic80_vtbl != NULL ) {
		vfree( __qic80_vtbl );
		TRACEx1( 4, "freed qic80_vtbl, %d bytes", 
			 SIZEOF(__qic80_vtbl) );
		__qic80_vtbl = NULL;
	}
	TRACE_EXIT;
}

/*
 * Jon says:
 *
 * __get_dma_pages() _will_ try to swap to make satisfy the request
 * but it doesn't doesn't have the intelligence to concentrate on
 * swapping contiguous DMAable ram, so may give up before finding
 * our request :-(
 * The below go straight for a 128k block, which should hopefully
 * leave us our 32k page somewhere :-)
 * 
 * Sorry, Jon, but Linux does not start swapping if it doesn't find a 
 * 128kb contigous block. The only way to get Linux starting swapping
 * seems to be to allocate small pages. That's ugly, but has some
 * probability to get us free pages.  I just allocate 4kb pages
 * up to a maximum number of (1 << (failures+3) )
 */

#ifdef ZFTAPE_SWAP_HACK
struct my_mem_node
{
	struct my_mem_node *next;
	unsigned long order;
};

static struct my_mem_node *
get_max_dma_block( int priority )
{
#if LINUX_VERSION_CODE < KERNEL_VER(1,3,74)
	int order = NR_MEM_LISTS;
#else
	int order = 6;
#endif
	struct my_mem_node *block;

	do {
		order --;
#if LINUX_VERSION_CODE < KERNEL_VER(1,3,0)
		block = (struct my_mem_node *)
			__get_dma_pages( priority, order );
#elif LINUX_VERSION_CODE < KERNEL_VER(1,3,74)
		block = (struct my_mem_node *)
			__get_dma_pages( priority, order );
#else
		block = (struct my_mem_node *)
			__get_dma_pages( priority, order );
#endif
		if ( block != NULL ) {
			block->order = order;
			break;
		}
	} while ( order > 0 );
	return block;
}

static struct my_mem_node *
get_max_block( int priority )
{
#if LINUX_VERSION_CODE < KERNEL_VER(1,3,74)
	int order = NR_MEM_LISTS;
#else
	int order = 6;
#endif
	struct my_mem_node *block;

	do {
		order --;
#if LINUX_VERSION_CODE < KERNEL_VER(1,3,0)
		block = (struct my_mem_node *)
			__get_free_pages( priority, order );
#elif LINUX_VERSION_CODE < KERNEL_VER(1,3,74)
		block = (struct my_mem_node *)
			__get_free_pages( priority, order, ~0UL );
#else
		block = (struct my_mem_node *)
			__get_free_pages( priority, order, 0 );
#endif
		if ( block != NULL ) {
			block->order = order;
			break;
		}
	} while ( order > 0 );
	return block;
}

#define MAX_ALLOC (high_memory / 2)

static int
make_linux_swap( int failures )
{
	TRACE_FUN( 5, "make_linux_swap");
	struct my_mem_node *mem_block, *list=NULL;
	unsigned long alloc_size;
	unsigned long allocated = 0; 
	int result = 0;

	TRACEi( 2, "try ", failures );
	alloc_size = BUFF_SIZE * (failures * 4);
	if ( alloc_size > MAX_ALLOC ) {
		TRACE( 2, "giving up" );
		TRACE_EXIT;
		return -ENOMEM;
	}
	TRACEx2( 2, "trying to allocate %ld bytes (max %ld)", 
		 alloc_size, MAX_ALLOC);
	ftape_sleep( 100*MILLISECOND );
	while ( allocated < alloc_size ) {
		mem_block = get_max_dma_block( GFP_KERNEL );
		schedule();
		if ( mem_block == NULL ) {
			mem_block = get_max_block( GFP_KERNEL );
			schedule();
		}
		if ( mem_block == NULL ) {
			TRACE(4,"get_max_block failed");
			result = -ENOMEM;
			break;
		}
		allocated += PAGE_SIZE * (1 << mem_block->order); 
		mem_block->next = list; 
		list            = mem_block;
		if (current->signal & _DONT_BLOCK) {
			TRACE( 2, "interrupted by non-blockable signal");
			result = -EINTR;
			break;
		}
	}
	while ( list != NULL ) {
		mem_block = list;
		list = list->next;
		free_pages( (unsigned long)mem_block, mem_block->order );
	}	
	TRACE_EXIT;
	return result;
}
#else
static int make_linux_swap( int failures ) { return -ENOMEM; }
#endif /* ZFTAPE_SWAP_HACK */

/*
 * Try to get num_buffers 32k buffers. As 32k is a power of 2 we
 * can get alligned 32k blocks dirrectly from the kernel.
 * If this proves to be as reliable as I hope we can call this at
 * open() rather than init_module() - Jon.
 *
 * Thanks, Jon. :-) Claus
 */
#define MAX_FAILURES 8 /* we try quite a lot to get the buffers */

/*
 *  try to allocate num_buffers dma buffers
 *
 *  returns the number of buffers currently allocated.
 */
#if defined(ALLOC_DMA_BUFFER_WHEN_OPENING_DEVICE) && defined(ZFTAPE_DMA_MEM_REDUCE_HACK)
static int
grow_number_of_buffers( int num_buffers )
{
	TRACE_FUN( 5, "grow_number_of_buffers");
	
	TRACEi( 7, "ftape_num_buffers: ", ftape_num_buffers );
	TRACEi( 7, "num_buffers: ", num_buffers );
	while ( ftape_num_buffers < num_buffers ) {
		void *new_buf;

		new_buf = (void *)vmalloc (BUFF_SIZE); 
		if ( new_buf != NULL ) {
			tape_buffers[ ftape_num_buffers ] = (char *)new_buf;
			TRACEx2( 3, "allocated buffer number %d at "
				 "address @ 0x%p", 
				 ftape_num_buffers, 
				 tape_buffers[ ftape_num_buffers ] );
			ftape_num_buffers++;
		} else {
			break;
		} 
	}
	TRACE_EXIT;
	return ftape_num_buffers;
}
#else
static int
grow_number_of_buffers( int num_buffers )
{
	TRACE_FUN( 5, "grow_number_of_buffers");
	int terminate = 0;
	
	TRACEi( 7, "ftape_num_buffers: ", ftape_num_buffers );
	TRACEi( 7, "num_buffers: ", num_buffers );
	while ( ftape_num_buffers < num_buffers && terminate < MAX_FAILURES ) {
		int failures = 0;
	        byte *new_buf;
		do {
#ifdef ZFTAPE_SWAP_HACK_TEST
			new_buf = NULL;
#else
			new_buf = dma_mem_alloc( BUFF_SIZE );
#endif
		} while ( (new_buf == NULL) 
			  && (make_linux_swap(failures++) == 0 ));
		if ( new_buf != 0 ) {
			tape_buffers[ ftape_num_buffers ] = new_buf;
			TRACEx2( 3, "allocated dma buffer number %d at "
				 "address @ 0x%p", 
				 ftape_num_buffers, 
				 tape_buffers[ ftape_num_buffers ] );
			ftape_num_buffers++;
		} else if (current->signal & _DONT_BLOCK) {
			terminate = MAX_FAILURES;
		} else {
			terminate ++;
			/*
			 *   should we ?
			 */
			ftape_sleep( 100*MILLISECOND );
		}
	}
	TRACE_EXIT;
	return ftape_num_buffers;
}
#endif
/*  The routine shrinks the number of dma-buffers, i.e. releases them.
 *  May be called with argument zero to get rid of all buffers 
 */
static int 
shrink_number_of_buffers( int num_buffers )
{
	TRACE_FUN( 5, "shrink_number_of_buffers");
	
	while ( ftape_num_buffers > num_buffers ) {
		ftape_num_buffers --;
#if defined(ALLOC_DMA_BUFFER_WHEN_OPENING_DEVICE) && defined(ZFTAPE_DMA_MEM_REDUCE_HACK)
		vfree( tape_buffers[ ftape_num_buffers ] );
#else
		free_pages( (unsigned long) tape_buffers[ ftape_num_buffers ],
			    __get_order( BUFF_SIZE) );
#endif 
		TRACEx2( 3, "freed tape buffer number %d at address @ 0x%p",
			 ftape_num_buffers, tape_buffers[ ftape_num_buffers ]);
		tape_buffers[ ftape_num_buffers ] = NULL;
	}
	TRACE_EXIT;
	return ftape_num_buffers;
}

/*  called by ftape_open() This function returns the number of dma buffers it 
 *  did allocate. Actually, it does allocate also some other buffers using 
 *  vmalloc(). Those buffers are also released when calling ftape_close()
 */
int
zftape_get_tape_buffers( kdev_t device_number )  
{
	TRACE_FUN( 5, "zftape_get_tape_buffer");
	int i = 0;
	int result;
#if defined(ALLOC_DMA_BUFFER_WHEN_OPENING_DEVICE) && defined(ZFTAPE_DMA_MEM_REDUCE_HACK)
	int failures = 0;

	do {
		dma_buffer = dma_mem_alloc(BUFF_SIZE);
	} while ( (dma_buffer == NULL) && (make_linux_swap(failures++) == 0 ));
	if ( dma_buffer == NULL ) {
		TRACE_EXIT;
		return 0;
	}
#endif
	if ( (result = grow_number_of_buffers ( desired_num_buffers )) 
	     >=
	     MIN_DMA_BUFFERS ) {
		for ( i = 0; i < NR_DYN_MEM ; i ++ ) {
			if ( dyn_mem_info[i].alloc_mask == 0 || 
			     (  MINOR( device_number ) 
			      & dyn_mem_info[i].alloc_mask) != 0 ) {
				*dyn_mem_info[i].location = 
					vmalloc( dyn_mem_info[i].size );
				if ( *dyn_mem_info[i].location == NULL ) {
					result = 0;
					break;
				}
				TRACEx3( 4,"allocated vmalloced buffer at %p, "
					 "size %ld, named %s", 
					 *dyn_mem_info[i].location, 
					 dyn_mem_info[i].size, 
					 dyn_mem_info[i].name );
			}
		}
	}
	if ( result == 0 )  {
		zftape_unget_tape_buffers();
	}
	TRACE_EXIT;
	return result;
}

/*    called by ftape_close() to get rid of all the temporary buffers, 
 *    including dma-memory.
 */
void
zftape_unget_tape_buffers( void )
{          
	TRACE_FUN( 5, "ftape_cleanup_mem");
	int i;
	
#if defined(ALLOC_DMA_BUFFER_WHEN_OPENING_DEVICE) && defined(ZFTAPE_DMA_MEM_REDUCE_HACK)
	if (dma_buffer) {
		free_pages((unsigned long)dma_buffer, __get_order(BUFF_SIZE) );
		dma_buffer = NULL; 
 	}
#endif
#ifdef ALLOC_DMA_BUFFER_WHEN_OPENING_DEVICE
	(void)shrink_number_of_buffers( 0 );
#else
	(void)shrink_number_of_buffers(MIN_DMA_BUFFERS);
#endif
	for ( i = 0; i < NR_DYN_MEM ; i ++ ) {
		if ( *dyn_mem_info[i].location ) {
			vfree ( *dyn_mem_info[i].location );
			TRACEx3( 4, "freed vmalloced buffer at %p, "
				 "size %ld, named %s", 
				 *dyn_mem_info[i].location, 
				 dyn_mem_info[i].size, dyn_mem_info[i].name );
			*dyn_mem_info[i].location = NULL;
		}
	}
	TRACE_EXIT;
}

/*
 *  called by MTSETDRVBUFFER to change the number of dma buffers
 *  The minimal number of dma-buffers for ftape to function satisfactory 
 *  is three, but MIN_DMA_BUFFERS may be specified in the Makefile.
 *  MAX_DMA_BUFFERS can also be specified using the Makefile
 *
 *  This routine is called using MTSETDRVBUFFER, that is, 
 *  "mt -f /dev/<your-favourite-zftape-device> mtsetdrvbuffer <multiple of 32 * 1024>
 *  changes the amount of dma-buffers the driver uses. A large number of buffers 
 *  speeds things up, but three or four buffers is enough for the driver.
 */
int
zftape_change_dma_size( int num_buffers )
{    
	TRACE_FUN( 5, "ftape_change_dma_size");
	
	if ( num_buffers > MAX_DMA_BUFFERS || num_buffers < MIN_DMA_BUFFERS ) {
		TRACEx3( 1,"number of buffers (%d) must be greater %d"
			 " and smaller %d", 
			 num_buffers, MIN_DMA_BUFFERS, MAX_DMA_BUFFERS );
		TRACE_EXIT;
		return -EINVAL;
	} else {
		desired_num_buffers = num_buffers;
		if ( desired_num_buffers <= ftape_num_buffers ) {
			(void)shrink_number_of_buffers ( desired_num_buffers );
			TRACE_EXIT;
			return 0;
		} else  {
			if ( grow_number_of_buffers ( desired_num_buffers ) 
			     < 
			     desired_num_buffers ) {
				TRACE_EXIT;
				return -ENOMEM;
			} else {
				TRACE_EXIT;
				return 0;
			}
		}
	}
}
